/* Copyright (C) 2020 Mark Sowden <markelswo@gmail.com>
 * Project Yin
 * */

#include "yin.h"
#include "map.h"
#include "gfx.h"
#include "act.h"
#include "game.h"

#define MAP_VERSION 10

typedef struct MapTexture {
	char textureName[ 12 ];
} MapTexture;

typedef struct MapVertex {
	int16_t x, y;
} MapVertex;

typedef struct MapLine {
	uint16_t startVertex;
	uint16_t endVertex;
	uint16_t flags;
	uint16_t unknown0;
	uint16_t colSomething;
	uint8_t  unknown1;
	uint8_t  hScale;
	uint16_t unknown2;
	uint32_t unknown3;
	uint16_t unknown4;

	PLVector2 normal;
} MapLine;

typedef struct MapArea {
	uint32_t     unknown0;
	uint16_t     unused0;
	uint16_t     unused1;
	unsigned int numLines;
	unsigned int *lineIndices;
	int          max[ 2 ]; /* boundary maximum */
	int          min[ 2 ]; /* boundary minimum */
} MapArea;

static struct {
	char paletteName[ 12 ];
	char skyName[ 12 ];
	char mapName[ 8 ];
	char nextMapName[ 8 ];
	char bonusMapName[ 8 ];

	char missionDescriptions[ 50 ][ 4 ];

	unsigned int keyCardRequirement;
	unsigned int cdTrackNum;

	unsigned int gravity;
	unsigned int friction;

	char polygonTextures[ 8 ][ 6 ];

	unsigned char lightLevel;

	MapTexture   *textures;
	unsigned int numTextures;
	MapVertex    *vertices;
	size_t       numVertices;
	MapLine      *lines;
	size_t       numLines;
} mapData = {
		.textures = NULL,
		.numTextures = 0,
		.vertices = NULL,
		.numVertices = 0,
};

static void Map_LoadData( PLPackage *pkg ) {
	PLFile *filePtr = plLoadPackageFile( pkg, "MAP_DATA" );
	if( filePtr == NULL ) {
		PrintError( "Failed to find preference data!\nPL: %s\n", plGetError() );
	}

	if( plReadFile( filePtr, mapData.paletteName, 1, 12 ) != 12 ) {
		PrintError( "Failed to read in palette name!\nPL: %s\n", plGetError() );
	}

	if( plReadFile( filePtr, mapData.skyName, 1, 12 ) != 12 ) {
		PrintError( "Failed to read in sky name!\nPL: %s\n", plGetError() );
	}

	if( plReadFile( filePtr, mapData.mapName, 1, 8 ) != 8 ) {
		PrintError( "Failed to read in map name!\nPL: %s\n", plGetError() );
	}

	if( plReadFile( filePtr, mapData.nextMapName, 1, 8 ) != 8 ) {
		PrintError( "Failed to read in next map name!\nPL: %s\n", plGetError() );
	}

	if( plReadFile( filePtr, mapData.bonusMapName, 1, 8 ) != 8 ) {
		PrintError( "Failed to read in bonus map name!\nPL: %s\n", plGetError() );
	}

	plCloseFile( filePtr );
}

static void Map_LoadTextureList( PLPackage *pkg ) {
	PLFile *filePtr = plLoadPackageFile( pkg, "TEX_LIST" );
	if( filePtr == NULL ) {
		PrintError( "Failed to find texture data!\nPL: %s\n", plGetError() );
	}

	/* seek over the ident */
	plFileSeek( filePtr, 4, PL_SEEK_CUR );

	bool status;
	mapData.numTextures = plReadInt32( filePtr, false, &status );

	/* skip over the table offset */
	plFileSeek( filePtr, 4, PL_SEEK_CUR );

	mapData.textures = Sys_AllocateMemory( mapData.numTextures, sizeof( MapTexture ) );
	for( unsigned int i = 0; i < mapData.numTextures; ++i ) {
		if( plReadFile( filePtr, &mapData.textures[ i ], 1, 12 ) != 12 ) {
			PrintError( "Failed to read in texture index!\nPL: %s\n", plGetError() );
		}

		/* skip the other index data */
		plFileSeek( filePtr, 8, PL_SEEK_CUR );
	}

	plCloseFile( filePtr );
}

static void Map_LoadVertices( PLPackage *pkg ) {
	PLFile *filePtr = plLoadPackageFile( pkg, "VERTICES" );
	if ( filePtr == NULL ) {
		PrintError( "Failed to find vertex data!\nPL: %s\n", plGetError() );
	}

	/* figure out how many vertices are here */
	size_t lumpSize = plGetFileSize( filePtr );
	static const unsigned int vertexSize = 8;
	mapData.numVertices = lumpSize / vertexSize;

	bool status;
	mapData.vertices = Sys_AllocateMemory( mapData.numVertices, sizeof( MapVertex ) );
	for ( unsigned int i = 0; i < mapData.numVertices; ++i ) {
		/* flipped so they match up with what we need */
		mapData.vertices[ i ].y = ( int16_t ) ( plReadInt32( filePtr, false, &status ) >> 16 );
		mapData.vertices[ i ].x = ( int16_t ) ( plReadInt32( filePtr, false, &status ) >> 16 );
	}

	if( !status ) {
		PrintError( "Failed to load vertex data!\nPL: %s\n", plGetError() );
	}

	plCloseFile( filePtr );
}

static void Map_LoadLines( PLPackage *wad ) {
	PLFile *filePtr = plLoadPackageFile( wad, "LINES" );
	if ( filePtr == NULL ) {
		PrintError( "Failed to find line data!\nPL: %s\n", plGetError() );
	}

	/* figure out how many lines are here */
	size_t lumpSize = plGetFileSize( filePtr );
	static const unsigned int lineSize = 16;
	mapData.numLines = lumpSize / lineSize;

	bool status;
	mapData.lines = Sys_AllocateMemory( mapData.numLines, sizeof( MapLine ) );
	for ( unsigned int i = 0; i < mapData.numLines; ++i ) {
		mapData.lines[ i ].startVertex  = plReadInt16( filePtr, false, &status );
		if ( mapData.lines[ i ].startVertex >= mapData.numVertices ) {
			PrintError( "Invalid start vertex for line %d!\n", i );
		}

		mapData.lines[ i ].endVertex    = plReadInt16( filePtr, false, &status );
		if ( mapData.lines[ i ].endVertex >= mapData.numVertices ) {
			PrintError( "Invalid end vertex for line %d!\n", i );
		}

		/*
		mapData.lines[ i ].flags        = plReadInt16( filePtr, false, &status );
		mapData.lines[ i ].unknown0     = plReadInt16( filePtr, false, &status );
		mapData.lines[ i ].colSomething = plReadInt16( filePtr, false, &status );
		mapData.lines[ i ].unknown1     = plReadInt8( filePtr, &status );
		mapData.lines[ i ].hScale       = plReadInt8( filePtr, &status );
		mapData.lines[ i ].unknown2     = plReadInt16( filePtr, false, &status );
		mapData.lines[ i ].unknown3     = plReadInt32( filePtr, false, &status );
		mapData.lines[ i ].unknown4     = plReadInt16( filePtr, false, &status );
		*/

		plFileSeek( filePtr, 12, PL_SEEK_CUR );
	}

	if( !status ) {
		PrintError( "Failed to load line data!\nPL: %s\n", plGetError() );
	}

	plCloseFile( filePtr );
}

void Map_LoadAreas( PLPackage *wad ) {
#if 0
	PLFile *filePtr = plLoadPackageFile( wad, "M_AREAS" );
	if ( filePtr == NULL ) {
		PrintError( "Failed to load area data!\nPL: %s\n", plGetError() );
	}

	bool status;
	mapData.numAreas = plReadInt32( filePtr, false, &status );
	mapData.areas = Sys_AllocateMemory( mapData.numAreas, sizeof( MapArea ) );
	for ( unsigned int i = 0; i < mapData.numAreas; ++i ) {
		MapArea *area = &mapData.areas[ i ];
		area->unknown0 = plReadInt32( filePtr, false, &status );
		area->unused0  = plReadInt16( filePtr, false, &status );
		area->unused1  = plReadInt16( filePtr, false, &status );
		area->numLines = plReadInt16( filePtr, false, &status );

		/* generate a list of all our line indices */
		area->lineIndices = Sys_AllocateMemory( mapData.areas[ i ].numLines, sizeof( unsigned int ) );

		area->max[ 0 ] = area->max[ 1 ] = INT32_MIN;
		area->min[ 0 ] = area->min[ 1 ] = INT32_MAX;

		for ( unsigned int j = 0; j < mapData.areas[ i ].numLines; ++j ) {
			mapData.areas[ i ].lineIndices[ j ] = plReadInt16( filePtr, false, &status );
			if ( mapData.areas[ i ].lineIndices[ j ] >= mapData.numLines ) {
				PrintError( "Invalid line index %d in area %d!\n", mapData.areas[ i ].lineIndices[ j ], i );
			}

			MapLine *curLine = &mapData.lines[ mapData.areas[ i ].lineIndices[ j ] ];

			/* calculate the area bounds */
			const MapPoint *points[ 2 ];
			points[ 0 ] = &mapData.points[ curLine->startVertex ];
			points[ 1 ] = &mapData.points[ curLine->endVertex ];

			for ( unsigned int k = 0; k < 2; ++k ) {
				if ( points[ k ]->x > area->max[ 0 ] ) {
					area->max[ 0 ] = points[ k ]->x;
				}

				if ( points[ k ]->y > area->max[ 1 ] ) {
					area->max[ 1 ] = points[ k ]->y;
				}

				if ( points[ k ]->x < area->min[ 0 ] ) {
					area->min[ 0 ] = points[ k ]->x;
				}

				if ( points[ k ]->y < area->min[ 1 ] ) {
					area->min[ 1 ] = points[ k ]->y;
				}
			}

			/* generate the normal for this particular face */
			curLine->normal = plComputeLineNormal( &PLVector2( points[ 0 ]->x, points[ 0 ]->y ), &PLVector2( points[ 1 ]->x, points[ 1 ]->y ) );
		}
	}

	plCloseFile( filePtr );
#endif
}

bool Map_ValidateVersion( PLPackage *pkg ) {
	bool status;
	PLFile *filePtr = plLoadPackageFile( pkg, "VERSION" );
	unsigned int version = plReadInt32( filePtr, false, &status );
	plCloseFile( filePtr );

	if( !status ) {
		PrintWarn( "Failed to read in map version!\nPL: %s\n", plGetError() );
		return false;
	}

	return ( version == MAP_VERSION );
}

void Map_Load( const char *mapName ) {
	PrintMsg( "Loading %s...\n" );

	PLPackage *pkg = plLoadPackage( mapName );
	if( pkg == NULL ) {
		PrintError( "Failed to load map, \"%s\"!\nPL: %s\n", mapName, plGetError() );
	}

	/* go ahead and check the version */
	if( !Map_ValidateVersion( pkg ) ) {
		PrintError( "Invalid map version, aborting!\n" );
	}

	Map_LoadData( pkg );
	Map_LoadVertices( pkg );
	Map_LoadLines( pkg );
	Map_LoadTextureList( pkg );

	plDestroyPackage( pkg );
}

bool Map_CheckCollisions( const PLCollisionAABB *bounds, unsigned int curArea ) {
#if 0
	const MapArea *area = &mapData.areas[ curArea ];
	for( unsigned int j = 0; j < area->numLines; ++j ) {
		MapLine  *line = &mapData.lines[ area->lineIndices[ j ] ];
		MapPoint *startPoint = &mapData.points[ line->startVertex ];
		MapPoint *endPoint = &mapData.points[ line->endVertex ];

		//bool hit = plIsAABBIntersectingLine( bounds, &PLVector2( startPoint->x, startPoint->y ), &PLVector2( endPoint->x, endPoint->y ), &line->normal );
		float hitValue;
		bool hit = plIsPointIntersectingLine( &PLVector2( bounds->origin.x, bounds->origin.z ), &PLVector2( startPoint->x, startPoint->y ), &PLVector2( endPoint->x, endPoint->y ), &line->normal, &hitValue );
		if( hit ) {
			//PrintMsg( "HIT: %f\n", hitValue );
			//return true;
		}

		//PrintMsg( "NO HIT: %f\n", hitValue );
	}
#endif

	return false;
}

void Map_Draw( void ) {
	/* fetch the local player so we can perform vis testing */
	Actor *player = Gam_GetPlayer();
	if ( player == NULL ) {
		return;
	}

	Gfx_EnableShaderProgram( SHADER_LIT );

	for( unsigned int i = 0; i < mapData.numLines; ++i ) {
		const MapVertex *startPoint = &mapData.vertices[ mapData.lines[ i ].startVertex ];
		const MapVertex *endPoint = &mapData.vertices[ mapData.lines[ i ].endVertex ];

		/* in the long term this should obviously all just get batched... */
		plDrawTexturedQuad(
			&PLVector3( startPoint->x, 128.0f, startPoint->y ),
			&PLVector3( endPoint->x, 128.0f, endPoint->y ),
			&PLVector3( startPoint->x, 0, startPoint->y ),
			&PLVector3( endPoint->x, 0, endPoint->y ),
			4.0f, 4.0f,
			Gfx_GetFallbackTexture()
		);
	}

#if 0
	/* prototype only supports a single wall texture at a time */
	PLTexture *wallTexture = Gfx_GetWallTexture( 20 );
	unsigned int wallHeight = wallTexture->h * 2;

	for ( unsigned int i = 0; i < mapData.numAreas; ++i ) {
		const MapArea *area = &mapData.areas[ i ];
		for ( unsigned int j = 0; j < area->numLines; ++j ) {
			MapPoint *startPoint = &mapData.points[ mapData.lines[ area->lineIndices[ j ] ].startVertex ];
			MapPoint *endPoint = &mapData.points[ mapData.lines[ area->lineIndices[ j ] ].endVertex ];

			/* ensure the wall is visible before we draw it */
			bool aVisible = Player_IsPointVisible( player, &PLVector2( startPoint->x, startPoint->y ) );
			bool bVisible = Player_IsPointVisible( player, &PLVector2( endPoint->x, endPoint->y ) );
			if ( !aVisible && !bVisible ) {
				continue;
			}

			/* in the long term this should obviously all just get batched... */
			plDrawTexturedQuad(
					&PLVector3( startPoint->x, wallHeight, startPoint->y ),
					&PLVector3( endPoint->x, wallHeight, endPoint->y ),
					&PLVector3( startPoint->x, 0, startPoint->y ),
					&PLVector3( endPoint->x, 0, endPoint->y ),
					2, 2,
					wallTexture
			);

#ifdef DEBUG_WALL_NORMALS
			Gfx_EnableShaderProgram( SHADER_GENERIC );

			PLVector2 linePos;
			linePos = plAddVector2( PLVector2( startPoint->x, startPoint->y ), PLVector2( endPoint->x, endPoint->y ) );
			linePos = plDivideVector2f( &linePos, 2.0f );
			
			PLVector2 lineEndPos;
			lineEndPos = plAddVector2( linePos, plScaleVector2f( &mapData.lines[ area->lineIndices[ j ] ].normal, 64.0f ) );

			PLMatrix4 transform = plMatrix4Identity();
			plDrawSimpleLine( &transform, &PLVector3( linePos.x, 16.0f, linePos.y ), &PLVector3( lineEndPos.x, 16.0f, lineEndPos.y ), &PLColour( 255, 0, 0, 255 ) );

			Gfx_EnableShaderProgram( SHADER_LIT );
#endif
		}

		/* draw the ceiling and floor */
		plDrawTexturedQuad(
				&PLVector3( area->max[ 0 ], 0.0f, area->max[ 1 ] ),
				&PLVector3( area->min[ 0 ], 0.0f, area->max[ 1 ] ),
				&PLVector3( area->max[ 0 ], 0.0f, area->min[ 1 ] ),
				&PLVector3( area->min[ 0 ], 0.0f, area->min[ 1 ] ),
				2, 2,
				Gfx_GetFloorTexture( 0 )
		);
#ifndef DEBUG_CAM
		plDrawTexturedQuad(
				&PLVector3( area->max[ 0 ], wallHeight, area->max[ 1 ] ),
				&PLVector3( area->min[ 0 ], wallHeight, area->max[ 1 ] ),
				&PLVector3( area->max[ 0 ], wallHeight, area->min[ 1 ] ),
				&PLVector3( area->min[ 0 ], wallHeight, area->min[ 1 ] ),
				2, 2,
				Gfx_GetFloorTexture( 1 )
		);
#endif
	}
#endif
}
